// TestApp1.cpp : Defines the entry point for the application.
//

#ifdef _WIN32
#include "Win32/stdafx.h"
#endif

#include <algorithm>
#include <DK.h>


#define MAX_SIZE 50000000ULL

struct Value64
{
	int num;
	int dummy[15];
	bool operator < (const Value64& r) const
	{
		return this->num < r.num;
	}
};

struct Value16
{
	int num;
	int dummy[3];
	bool operator < (const Value16& r) const
	{
		return this->num < r.num;
	}
};

union Value8
{
	size_t i64;
	struct
	{
		uint32_t h, l;
	} i32;
	bool operator < (const Value8& r) const
	{
		return this->i64 < r.i64;
	}
};

using Value4 = uint32_t;


template <typename T, typename Rand>
void DoTest(bool dkFirst, Rand&& r)
{
	T* val1 = new T[MAX_SIZE];
	T* val2 = new T[MAX_SIZE];

	DKLog("Generating data... (%s)", typeid(T).name());
	for (size_t i = 0; i < MAX_SIZE; ++i)
		val1[i] = val2[i] = r();

	
	DKThread::Sleep(0.2);

	auto testSTL = [](T* val, size_t len)->DKHashResultSHA1
	{
		DKLogI("Sorting by std::sort...");
		DKTimer timer;
		timer.Reset();
		std::sort(val, val + len, [](const auto& lhs, const auto& rhs)->bool
		{
			return lhs < rhs;
		});
		double e1 = timer.Elapsed();
		DKLogW("std::sort: %f", e1);
		auto hash = DKHashSHA1(val, sizeof(T) * len);
		DKLog("std::sort hash: %ls", (const wchar_t*)hash.String());
		return hash;
	};

	auto testDK = [](T* val, size_t len)->DKHashResultSHA1
	{
		DKLogI("Sorting by DKStaticArray::Sort...");
		DKStaticArray<T> tmp(val, len);
		DKTimer timer;
		timer.Reset();
		tmp.Sort([](const auto& lhs, const auto& rhs)->bool
		{
			return lhs < rhs;
		});
		double e2 = timer.Elapsed();
		DKLogW("DKStaticArray::Sort: %f (SwapMethod:%d)", e2, tmp.SwapMethod);
		auto hash = DKHashSHA1(val, sizeof(T) * len);
		DKLog("DKStaticArray::Sort hash: %ls", (const wchar_t*)hash.String());
		return hash;
	};

	if (dkFirst)
	{
		auto hash1 = testDK(val1, MAX_SIZE);
		auto hash2 = testSTL(val2, MAX_SIZE);
		DKLog("Hash compare: %d", hash1.Compare(hash2));
	}
	else
	{
		auto hash1 = testSTL(val1, MAX_SIZE);
		auto hash2 = testDK(val2, MAX_SIZE);
		DKLog("Hash compare: %d", hash1.Compare(hash2));
	}

	delete[] val1;
	delete[] val2;
	DKLog("--------------");
	DKThread::Sleep(0.2);
}

#ifdef _WIN32
int APIENTRY wWinMain(_In_ HINSTANCE hInstance,
	_In_opt_ HINSTANCE hPrevInstance,
	_In_ LPWSTR    lpCmdLine,
	_In_ int       nCmdShow)
#else
int main(int argc, const char * argv[])
#endif
{
	DKApplication app;

	DoTest<Value64>(true, []()->Value64
	{
		Value64 val;
		val.num = DKRandom();
		return val;
	});
	DoTest<Value16>(true, []()->Value16
	{
		Value16 val;
		val.num = DKRandom();
		return val;
	});
	DoTest<Value8>(true, []()->Value8
	{
		return { size_t(DKRandom()) | (size_t(DKRandom()) << 32) };
	});
	DoTest<Value4>(true, []()->Value4
	{
		return DKRandom();
	});

	return 0;
}
